#!/usr/bin/python

import optparse

from launchpadlib.launchpad import Launchpad

lp = Launchpad.login_with("package-import", service_root="production")

lp_distro_releases = {
    "ubuntu": ["warty", "hoary", "breezy", "dapper", "edgy", "feisty",
               "gutsy", "hardy", "intrepid", "jaunty", "karmic", "lucid",
               "maverick", "natty", "oneiric",],
    "debian": ["lenny", "sid", "experimental", "squeeze"],
}

distro_pockets = {"ubuntu": ["Release", "Security", "Proposed", "Updates", "Backports"],
    "debian": ["Release",],
}


def make_suite(release, pocket):
    if pocket == "Release":
        suite = release
    elif pocket in ("Security", "Proposed", "Updates", "Backports"):
        suite = "%s-%s" % (release, pocket.lower())
    else:
        assert False, "Unkown pocket: %s" % pocket
    return suite


parser = optparse.OptionParser()
options, args = parser.parse_args()


def delete_series(distribution, series, package):
    for pocket in distro_pockets[distribution_name]:
        suite = make_suite(series.name, pocket)
        unique_name = "~ubuntu-branches/%s/%s/%s/%s" % (distribution.name, series.name, package, suite)
        branch = lp.branches.getByUniqueName(
            unique_name=unique_name)
        if branch is None:
            #print "Couldn't find branch at %s" % unique_name
            continue
        sp = series.getSourcePackage(name=package)
        other_branch = sp.getBranch(pocket=pocket)
        if other_branch == branch:
            print "Unlinking %s from series" % unique_name
            sp.setBranch(branch=None, pocket=pocket)
        for bug in branch.linked_bugs:
            print "Unlinking %s from %s" % (bug.id, unique_name)
            branch.unlinkBug(bug=bug)
        if not branch.canBeDeleted():
            print "%s can't be deleted" % unique_name
            continue
        print "Deleting %s" % branch.self_link
        branch.lp_delete()

for package in args:
    for distribution_name in ("debian", "ubuntu"):
        distribution = lp.distributions[distribution_name]
        current_series = distribution.current_series
        for series in reversed(distribution.series):
            delete_series(distribution, series, package)
