import types


from bzrlib import (
    config,
    osutils,
    )

# Directory containing all python scripts and the configuration file
_scripts_dir = osutils.dirname(osutils.realpath(__file__))

def _build_path(name):
    """Build and return a path in the scripts directory for name"""
    return osutils.pathjoin(_scripts_dir, name)


def pkgimportconfig_file_name():
    return _build_path('pkgimport.conf')


# FIXME: We don't inherit from LockableConfig because we don't want to create a
# 'lock' dir alonside the scripts (we should not require write access there and
# we don't need to modify options in the configuration file) -- vila 2011-03-22
class PkgimportConfig(config.IniBasedConfig):

    def __init__(self, file_name=None):
        if file_name is None:
            file_name = pkgimportconfig_file_name()
        super(PkgimportConfig, self).__init__(file_name=file_name)

    def config_id(self):
        return 'pkgimport'

    def get(self, name):
        return self.get_user_option(name)


def compatible_get(self, name, default_value=None):
    value = self.get_user_option(name)
    if value is None:
        value = default_value
    return value


class ConfigStack(object):
    """A stack of configurations where an option can be defined"""

    def __init__(self, config_list):
        self.list = []
        for c in config_list:
            if hasattr(c, 'get_user_option') and not hasattr(c, 'get'):
                c.get = types.MethodType(compatible_get, c, c.__class__)
            if not hasattr(c, 'get'):
                raise AssertionError("%r does not provide a 'get' method"
                                     % (c,))
            self.list.append(c)

    def get(self, name):
        for c in self.list:
            value = c.get(name)
            if value is not None:
                break
        return value


class Iconfig(ConfigStack):
    """An import packager configuration taking ``locations.conf`` into account.

    This allows one to override the values in ``pkgimport.conf`` in
    ``locations.conf`` for specified sections.
    """

    def __init__(self, location=None):
        if location is None:
            location = _scripts_dir
        super(Iconfig, self).__init__(
            [config.LocationConfig(location), PkgimportConfig()])

