#!/usr/bin/python
"""Invoke logrotate for the package importer taking configuration into account.
"""

import os
import subprocess
import sys

import iconfig

logrotate_configuration='''%s {
    daily
    dateext
    rotate 7
    delaycompress
    compress
    notifempty
    missingok
}

%s {
    daily
    dateext
    rotate 28
    delaycompress
    compress
    notifempty
    missingok
}
'''

def main():
    conf = iconfig.Iconfig()
    debug_log = conf.get('pkgimport.driver.log.debug')
    progress_log = conf.get('pkgimport.driver.log.progress')

    log_dir = conf.get('pkgimport.driver.log_dir')
    lrot_conf_path = os.path.join(log_dir, 'logrotate.conf')
    # We let the open fail if the directory doesn't exist, either there is no
    # log files or the setup is misconfigured anyway.
    lrot_conf = open(lrot_conf_path, 'w')
    try:
        # FIXME: Another use case for config.expand_options -- vila 2011-03-24
        lrot_conf.write(logrotate_configuration % (debug_log, progress_log))
    finally:
        lrot_conf.close()

    lrot_state_path = os.path.join(log_dir, 'logrotate.state')
    retcode = subprocess.call(['/usr/sbin/logrotate', '-s', lrot_state_path,
                               lrot_conf_path])
    return retcode


if __name__ == '__main__':
    sys.exit(main())
