#!/usr/bin/python

import os
import sys
import unittest


class TestLoader(unittest.TestLoader):
    """Custom TestLoader to extend the stock python one."""

    suiteClass = unittest.TestSuite

    def loadTestsFromModuleNames(self, names):
        result = self.suiteClass()
        for name in names:
            result.addTests(self.loadTestsFromModuleName(name))
        return result

    def loadTestsFromModuleName(self, name):
        result = self.suiteClass()
        __import__(name, globals(), locals(), [])
        module = sys.modules[name]

        result.addTests(self.loadTestsFromModule(module))
        return result


class TestProgram(unittest.TestProgram):

    def __init__(self, module='__main__'):
        # Tell unittest to call createTests
        unittest.TestProgram.__init__(self, module, testLoader=TestLoader(),
                                      # Tell unittest to call createTests
                                      defaultTest=True,
                                      )

    def createTests(self):
        suite = self.testLoader.suiteClass()

        std_tests = self.testLoader.loadTestsFromModule(self.module)
        suite.addTests(std_tests)
        test_module_names = [
            'config',
            'import_list',
            'import_package',
            'mass_import',
            'package_to_import',
            'revid_database',
            'status_database',
            ]
        suite.addTests(self.testLoader.loadTestsFromModuleNames(
                ['tests']
                + ['tests.test_%s' % (name,) for name in test_module_names]))
        self.test = suite


if __name__ == '__main__':
    TestProgram(module=__name__)


