from bzrlib import (
    config,
    osutils,
    tests,
    )


import iconfig


class TestConfigPath(tests.TestCase):

    def setUp(self):
        super(TestConfigPath, self).setUp()
        self.mydir = osutils.dirname(osutils.realpath(iconfig.__file__))

    def test_default_path(self):
        # Ensure the iconfig finds the right path
        self.assertEquals(self.mydir, iconfig._scripts_dir)

    def test_iconfig_file_name(self):
        conf_path = iconfig.pkgimportconfig_file_name()
        self.assertEquals(self.mydir, osutils.dirname(conf_path))
        self.assertEndsWith(conf_path, 'pkgimport.conf')


class TestPkgimportconfig(tests.TestCaseInTempDir):

    def setUp(self):
        super(TestPkgimportconfig, self).setUp()
        self.conf_file_name = osutils.pathjoin(self.test_dir, 'pkgimport.conf')

    def test_simple_option(self):
        conf = iconfig.PkgimportConfig.from_string(
            '''pkgimport.base_dir=/foo/bar''',
            file_name=self.conf_file_name, save=True)
        self.assertEquals('/foo/bar', conf.get('pkgimport.base_dir'))


class TestDefaultPkgimportconfig(tests.TestCase):

    def setUp(self):
        super(TestDefaultPkgimportconfig, self).setUp()
        self.conf = iconfig.PkgimportConfig()

    def test_default_config(self):
        self.assertEquals('pkgimport', self.conf.config_id())
        self.assertEquals(iconfig.pkgimportconfig_file_name(),
                          self.conf.file_name)

    def test_default_basedir(self):
        self.assertEquals('/srv/package-import.canonical.com/new',
                          self.conf.get('pkgimport.base_dir'))

    def assertOptiondIsDefined(self, name):
        self.assertIsNot(None, self.conf.get(name))

    def test_required_options(self):
        self.assertOptiondIsDefined('pkgimport.base_dir')
        self.assertOptiondIsDefined('pkgimport.log_dir')
        self.assertOptiondIsDefined('pkgimport.driver.log_dir')
        self.assertOptiondIsDefined('pkgimport.driver.log.debug')
        self.assertOptiondIsDefined('pkgimport.driver.log.progress')
        self.assertOptiondIsDefined('pkgimport.packages.log_dir')


class TestConfigStack(tests.TestCase):

    def test_compatibility(self):
        self.assertRaises(AssertionError, iconfig.ConfigStack, [object()])

    def test_single_config_get(self):
        conf = config.IniBasedConfig.from_string('foo=bar')
        conf_stack = iconfig.ConfigStack([conf])
        self.assertEquals('bar', conf_stack.get('foo'))

    def test_get_first_definition(self):
        conf1 = config.IniBasedConfig.from_string('foo=bar')
        conf2 = config.IniBasedConfig.from_string('foo=baz')
        conf_stack = iconfig.ConfigStack([conf1, conf2])
        self.assertEquals('bar', conf_stack.get('foo'))


class TestIconfig(tests.TestCaseInTempDir):

    def test_location_override_default(self):
        lconf = config.LocationConfig('.')
        lconf.set_user_option('pkgimport.base_dir', 'foo')
        conf = iconfig.Iconfig('.')
        self.assertEquals('foo', conf.get('pkgimport.base_dir'))
