from bzrlib import tests

try:
    from debian import changelog
except ImportError:
    from debian_bundle import changelog


import icommon
from icommon import PackageToImport as PTI
V = changelog.Version


class PackageToImportTests(tests.TestCase):

    def test_same_as_same(self):
        a = PTI("a", V("1"), "debian", "sid", "release",
                component="main", on_lp=True, url=None)
        b = PTI("a", V("1"), "debian", "sid", "release", component="contrib")
        c = PTI("a", V("1"), "debian", "sid", "release", on_lp=False)
        d = PTI("a", V("1"), "debian", "sid", "release", url="a")
        self.assertTrue(a.same_as(b))
        self.assertTrue(a.same_as(c))
        self.assertTrue(a.same_as(d))

    def test_same_as_different(self):
        a = PTI("a", V("1"), "debian", "sid", "release",
                component="main", on_lp=True, url=None)
        b = PTI("b", V("1"), "debian", "sid", "release")
        c = PTI("a", V("1.1"), "debian", "sid", "release")
        d = PTI("a", V("1"), "ubuntu", "sid", "release")
        e = PTI("a", V("1"), "debian", "squeeze", "release")
        f = PTI("a", V("1"), "debian", "sid", "security")
        self.assertFalse(a.same_as(b))
        self.assertFalse(a.same_as(c))
        self.assertFalse(a.same_as(d))
        self.assertFalse(a.same_as(e))
        self.assertFalse(a.same_as(f))

    def test_get_url(self):
        a = PTI("a", V("1"), "debian", "sid", "release", url="url")
        self.assertEqual("url", a.get_url())
        b = PTI("a", V("1"), "debian", "sid", "release",
                component="non-free", on_lp=False)
        self.assertEqual(icommon.urllib_debian_base_url
                + "pool/non-free/a/a/a_1.dsc", b.get_url())
        c = PTI("libfoo", V("1:1.1-1"), "debian", "squeeze", "release",
                component="contrib", on_lp=False)
        self.assertEqual(icommon.urllib_debian_base_url
                + "pool/contrib/libf/libfoo/libfoo_1.1-1.dsc", c.get_url())
        c = PTI("libfoo", V("1:1.1-1"), "debian", "squeeze", "release",
                component="contrib", on_lp=True)
        self.assertEqual(icommon.urllib_launchpad_base_url
                + "debian/squeeze/+source/libfoo/1:1.1-1/+files/libfoo_1.1-1.dsc",
                c.get_url())
        d = PTI("libfoo", V("1:1.1-1"), "ubuntu", "lucid", "security")
        self.assertEqual(icommon.urllib_launchpad_base_url
                + "ubuntu/lucid/+source/libfoo/1:1.1-1/+files/libfoo_1.1-1.dsc",
                d.get_url())

    def test_str(self):
        a = PTI("a", V("1"), "debian", "sid", "release",
                component="main", on_lp=True, url=None)
        self.assertEqual("PackageToImport(a, 1, debian, sid, release, main)",
                str(a))

    def assertOrdering(self, a, b):
        self.assertTrue(PTI.import_order_comparator(a, b) < 0)

    def test_comparison(self):
        # Distro is the most-significant factor, with all of Ubuntu coming
        # after all of Debian, irregardless of releases and versions:
        self.assertOrdering(PTI("a", V("999999"), "debian", "sid", "release"),
                            PTI("a", V("1"), "ubuntu", "warty", "release"))

        # Within a distro, release is the next most significant:
        self.assertOrdering(PTI("a", V("1"), "debian", "lenny", "release"),
                            PTI("a", V("1"), "debian", "squeeze", "release"))

        # Except for Ubuntu backports, which are a special case, imported after
        # all other Ubuntu imports in the same per-package run:
        self.assertOrdering(PTI("a", V("1"), "ubuntu", "natty", "release"),
                            PTI("a", V("1~hardy1"), "ubuntu", "hardy",
                                "backports"))

        # Within Ubuntu, pockets matter:
        self.assertOrdering(PTI("a", V("1"), "ubuntu", "natty", "release"),
                            PTI("a", V("1"), "ubuntu", "natty", "security"))
        self.assertOrdering(PTI("a", V("1"), "ubuntu", "natty", "security"),
                            PTI("a", V("1"), "ubuntu", "natty", "proposed"))
        self.assertOrdering(PTI("a", V("1"), "ubuntu", "natty", "proposed"),
                            PTI("a", V("1"), "ubuntu", "natty", "updates"))
        self.assertOrdering(PTI("a", V("1"), "ubuntu", "natty", "updates"),
                            PTI("a", V("1"), "ubuntu", "natty", "backports"))
