from bzrlib import tests


try:
    from debian import changelog
except ImportError:
    from debian_bundle import changelog


import icommon


class RevidDatabaseTests(tests.TestCaseWithTransport):

    def setUp(self):
        super(RevidDatabaseTests, self).setUp()
        self.cleanup = 0
        self.push = 0
        self.db = self.get_db()
        self.version = changelog.Version("1.0-1")
        self.version2 = changelog.Version("1.0-2")
        self.version3 = changelog.Version("1.0-3")
        self.suite = "karmic"
        self.tree = self.make_branch_and_tree(".")
        self.revid = self.tree.commit("one")
        self.revid2 = self.tree.commit("two")
        self.revid3 = self.tree.commit("three")

    def get_db(self):
        return icommon.RevidDatabase(":memory:", "pkg")

    def check_rows(self, memory, working, saved):
        self.assertEqual(memory, len(self.db._memory_rows()))
        self.assertEqual(working, len(self.db._working_rows()))
        self.assertEqual(saved, len(self.db._saved_rows()))

    def test_mark(self):
        self.assertFalse(self.db.is_marked(self.version, self.suite))
        self.check_rows(0, 0, 0)
        self.db.mark(self.version, self.revid, self.suite, self.tree.branch)
        self.check_rows(1, 0, 0)
        self.assertTrue(self.db.is_marked(self.version, self.suite))

    def test_mark_again(self):
        self.db.mark(self.version, self.revid, self.suite, self.tree.branch)
        self.db.commit_outstanding()
        self.db.mark(self.version, self.revid, self.suite, self.tree.branch)
        self.assertRaises(AssertionError, self.db.commit_outstanding)
        self.assertRaises(AssertionError, self.db.is_marked, self.version,
                self.suite)

    def test_check(self):
        self.db.mark(self.version, self.revid, self.suite, self.tree.branch)
        self.db.commit_outstanding()
        self.assertTrue(self.db.check(self.version, self.revid, self.suite,
                    self.tree.branch))
        self.assertFalse(self.db.check(changelog.Version(
                        str(self.version)+"1"),
                    self.revid, self.suite, self.tree.branch))
        self.assertFalse(self.db.check(self.version, self.revid,
                    self.suite+"a", self.tree.branch))

    def test_commit(self):
        self.db.mark(self.version, self.revid, self.suite, self.tree.branch)
        self.db.commit_outstanding()
        self.db.commit()
        self.check_rows(0, 0, 1)
        self.db.mark(self.version, self.revid, self.suite, self.tree.branch)
        self.assertRaises(AssertionError, self.db.commit_outstanding)
        self.check_rows(1, 0, 1)

    def do_cleanup(self):
        self.cleanup += 1

    def do_push(self):
        self.push += 1

    def test_cleanup_nothing_to_do(self):
        self.db.mark(self.version, self.revid, self.suite, self.tree.branch)
        self.db.commit()
        self.db.cleanup_last_run(self.do_cleanup, self.do_push)
        self.assertEqual(0, self.push)
        self.assertEqual(1, self.cleanup)

    def test_cleanup(self):
        self.db.mark(self.version, self.revid, self.suite, self.tree.branch)
        self.db.commit_outstanding()
        self.db.commit()
        self.db.mark(self.version2, self.revid2, self.suite,
                self.tree.branch)
        self.db.commit_outstanding()
        self.check_rows(0, 1, 1)
        self.db.cleanup_last_run(self.do_cleanup, self.do_push)
        self.assertEqual(1, self.push)
        self.assertEqual(1, self.cleanup)
        self.check_rows(0, 0, 2)

    def test_discard(self):
        self.db.mark(self.version, self.revid, self.suite, self.tree.branch)
        self.db.mark(self.version2, self.revid2, self.suite,
                self.tree.branch)
        self.db.commit_outstanding()
        self.db.commit()
        self.db.mark(self.version3, self.revid3, self.suite,
                self.tree.branch)
        self.db.commit_outstanding()
        self.check_rows(0, 1, 2)
        self.db.discard_last_run()
        self.check_rows(0, 0, 2)


