from bzrlib import tests


import icommon


class StatusDatabaseTests(tests.TestCase):

    def setUp(self):
        super(StatusDatabaseTests, self).setUp()
        self.db = icommon.StatusDatabase(":memory:")

    def test_no_failures_in_empty_db(self):
        reasons, package_info = self.db.summarise_failures()
        self.assertEquals({}, reasons)
        self.assertEquals([], package_info)

    def test_no_unemailed_failures_in_empty_db(self):
        self.assertEquals([], self.db.unemailed_failures())

    def assertSignature(self, expected, raw):
        self.assertEquals(expected, self.db.failure_signature(raw))

    def test_running_signature(self):
        # Running imports use a special failure signature
        self.assertSignature(None, icommon.running_sentinel)

    def test_one_line_signature(self):
        self.assertSignature('AssertionError', 'AssertionError: xx')

    def test_multi_line_signature(self):
        # We use a real-life example here, the apparent duplication of method
        # names is due to calls to the base class methods
        self.assertSignature(
            'OperationalError:do_one_step:do_one_step'
            ':collect_terminated_threads'
            ':collect_terminated_threads:collect'
            ':finish_job:_set_failure',
            '''Traceback (most recent call last):
  File "/srv/package-import.canonical.com/new/scripts/mass_import.py", line 234, in do_one_step
    super(ImportDriver, self).do_one_step()
  File "/srv/package-import.canonical.com/new/scripts/icommon.py", line 2091, in do_one_step
    self.collect_terminated_threads()
  File "/srv/package-import.canonical.com/new/scripts/mass_import.py", line 253, in collect_terminated_threads
    super(ImportDriver, self).collect_terminated_threads()
  File "/srv/package-import.canonical.com/new/scripts/icommon.py", line 2119, in collect_terminated_threads
    t.collect()
  File "/srv/package-import.canonical.com/new/scripts/mass_import.py", line 162, in collect
    unicode_output.encode("utf-8", "replace"))
  File "/srv/package-import.canonical.com/new/scripts/icommon.py", line 552, in finish_job
    self._set_failure(c, package, output, now)
  File "/srv/package-import.canonical.com/new/scripts/icommon.py", line 485, in _set_failure
    % self.FAILURES_TABLE, (package, reason, now))
OperationalError: table failures has 4 columns but 3 values were supplied
''')

    def test_regular_traceback(self):
        self.assertSignature(
                'bzrlib.errors.NoSuchTag:lookup_tag',
                                 '''Traceback (most recent call last):
  File "/usr/lib/python2.5/site-packages/bzrlib/tag.py", line 109, in lookup_tag
    raise errors.NoSuchTag(tag_name)
bzrlib.errors.NoSuchTag: No such tag: upstream-4.6.2
''')

    def test_traceback_with_verbose_execption(self):
        # make sure we get bzrlib.errors.NoFinalPath without being tricked by:
        # the final empty line, nor the 'file-id:' and 'root trans-id' lines.
        self.assertSignature(
            'bzrlib.errors.NoFinalPath:get_path:_determine_path:final_name',
            ''''Traceback (most recent call last):
  File "/usr/lib/python2.5/site-packages/bzrlib/transform.py", line 2368, in get_path
    self._known_paths[trans_id] = self._determine_path(trans_id)
  File "/usr/lib/python2.5/site-packages/bzrlib/transform.py", line 2358, in _determine_path
    name = self.transform.final_name(trans_id)
  File "/usr/lib/python2.5/site-packages/bzrlib/transform.py", line 470, in final_name
    raise NoFinalPath(trans_id, self)
bzrlib.errors.NoFinalPath: No final name for trans_id 'new-18'
file-id: None
root trans-id: 'new-0'

''')



